@@ -20,11 +20,14 @@
   THE SOFTWARE.
 */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <dirent.h>
 #include <unistd.h>
 #include <errno.h>
@@ -46,12 +49,6 @@
 #include "data.h"
 #include "ident.h"
 
-#ifdef NEED_SNPRINTF
-#undef SCOPE
-#define SCOPE static
-#include "snprintf.c"
-#endif
-
 #define NPREFIX 1024
 
 #ifndef MAXFONTFILENAMELEN
@@ -61,7 +58,12 @@
 #define MAXFONTNAMELEN 1024
 #endif
 
-static char *encodings_array[] =
+/* Two levels of macro calls are needed so that we stringify the value
+   of MAXFONT... and not the string "MAXFONT..." */
+#define QUOTE(x)	#x
+#define STRINGIFY(x)	QUOTE(x)
+
+static const char *encodings_array[] =
     { "ascii-0",
       "iso8859-1", "iso8859-2", "iso8859-3", "iso8859-4", "iso8859-5",
       "iso8859-6", "iso8859-6.8", "iso8859-6.8x", "iso8859-6.16",
@@ -80,20 +82,20 @@ static char *encodings_array[] =
       "gb2312.1980-0", "gb18030.2000-0", "gb18030.2000-1",
       "ksc5601.1987-0", "ksc5601.1992-3"};
 
-static char *extra_encodings_array[] =
+static const char *extra_encodings_array[] =
     { "iso10646-1", "adobe-fontspecific", "microsoft-symbol" };
 
 static ListPtr encodings, extra_encodings;
-static char *outfilename;
+static const char *outfilename;
 
 #define countof(_a) (sizeof(_a)/sizeof((_a)[0]))
 
-static int doDirectory(char*, int, ListPtr);
+static int doDirectory(const char*, int, ListPtr);
 static int checkEncoding(FT_Face face, char *encoding_name);
 static int checkExtraEncoding(FT_Face face, char *encoding_name, int found);
 static int find_cmap(int type, int pid, int eid, FT_Face face);
-static char* notice_foundry(char *notice);
-static char* vendor_foundry(signed char *vendor);
+static const char* notice_foundry(const char *notice);
+static const char* vendor_foundry(const signed char *vendor);
 static int readFontScale(HashTablePtr entries, char *dirname);
 ListPtr makeXLFD(char *filename, FT_Face face, int);
 static int readEncodings(ListPtr encodings, char *dirname);
@@ -114,7 +116,7 @@ static char *exclusionSuffix;
 static void
 usage(void)
 {
-    fprintf(stderr, 
+    fprintf(stderr,
             "mkfontscale [ -b ] [ -s ] [ -o filename ] [-x suffix ]\n"
             "            [ -a encoding ] [ -f fuzz ] [ -l ] "
             "            [ -e directory ] [ -p prefix ] [ -n ] [ -r ] \n"
@@ -144,7 +146,7 @@ main(int argc, char **argv)
 
     encodings = makeList(encodings_array, countof(encodings_array), NULL, 0);
 
-    extra_encodings = makeList(extra_encodings_array, 
+    extra_encodings = makeList(extra_encodings_array,
                                countof(extra_encodings_array),
                                NULL, 0);
     doBitmaps = 0;
@@ -205,7 +207,7 @@ main(int argc, char **argv)
             argn++;
         } else if(strcmp(argv[argn], "-U") == 0) {
             doISO10646_1_encoding = 1;
-            argn++;            
+            argn++;
         } else if(strcmp(argv[argn], "-s") == 0) {
             doScalable = 0;
             argn++;
@@ -245,7 +247,7 @@ main(int argc, char **argv)
     if(outfilename == NULL) {
         if(doBitmaps)
             outfilename = "fonts.dir";
-        else 
+        else
             outfilename = "fonts.scale";
     }
 
@@ -314,9 +316,9 @@ getName(FT_Face face, int nid)
     char *string;
     int i;
 
-    if(getNameHelper(face, nid, 
+    if(getNameHelper(face, nid,
                      TT_PLATFORM_MICROSOFT, TT_MS_ID_UNICODE_CS, &name) ||
-       getNameHelper(face, nid, 
+       getNameHelper(face, nid,
                      TT_PLATFORM_APPLE_UNICODE, -1, &name)) {
         string = malloc(name.string_len / 2 + 1);
         if(string == NULL) {
@@ -349,7 +351,7 @@ getName(FT_Face face, int nid)
     return NULL;
 }
 
-static char*
+static const char*
 os2Weight(int weight)
 {
     if(weight < 150)
@@ -368,11 +370,11 @@ os2Weight(int weight)
         return "bold";
     else if(weight < 850)
         return "extrabold";
-    else 
+    else
         return "black";
 }
 
-static char*
+static const char*
 os2Width(int width)
 {
     if(width <= 1)
@@ -395,15 +397,15 @@ os2Width(int width)
         return "ultraexpanded";
 }
 
-static char *widths[] = {
+static const char *widths[] = {
     "ultracondensed", "extracondensed", "condensed", "semicondensed",
-    "normal", "semiexpanded", "expanded", "extraexpanded", "ultraexpanded" 
+    "normal", "semiexpanded", "expanded", "extraexpanded", "ultraexpanded"
 };
 
 #define NUMWIDTHS (sizeof(widths) / sizeof(widths[0]))
 
-static char*
-nameWidth(char *name)
+static const char*
+nameWidth(const char *name)
 {
     char buf[500];
     int i;
@@ -420,13 +422,15 @@ nameWidth(char *name)
     return NULL;
 }
 
-static char*
-t1Weight(char *weight)
+static const char*
+t1Weight(const char *weight)
 {
     if(!weight)
         return NULL;
     if(strcmp(weight, "Thin") == 0)
         return "thin";
+    if(strcmp(weight, "ExtraLight") == 0) /* FontForge uses this for 200*/
+        return "extralight";
     if(strcmp(weight, "Light") == 0)
         return "light";
     if(strcmp(weight, "Regular") == 0)
@@ -447,6 +451,8 @@ t1Weight(char *weight)
         return "semibold";
     else if(strcmp(weight, "Bold") == 0)
         return "bold";
+    else if(strcmp(weight, "Heavy") == 0) /* FontForge uses this for 800*/
+        return "extrabold";
     else if(strcmp(weight, "Black") == 0)
         return "black";
     else {
@@ -458,13 +464,13 @@ t1Weight(char *weight)
 static int
 unsafe(char c)
 {
-    return 
+    return
         c < 0x20 || c > 0x7E ||
         c == '[' || c == ']' || c == '(' || c == ')' || c == '\\' || c == '-';
 }
 
-static char *
-safe(char* s)
+static const char *
+safe(const char* s)
 {
     int i, len, safe_flag = 1;
     char *t;
@@ -499,7 +505,7 @@ ListPtr
 makeXLFD(char *filename, FT_Face face, int isBitmap)
 {
     ListPtr xlfd = NULL;
-    char *foundry, *family, *weight, *slant, *sWidth, *adstyle, 
+    const char *foundry, *family, *weight, *slant, *sWidth, *adstyle,
         *spacing, *full_name;
     TT_Header *head;
     TT_HoriHeader *hhea;
@@ -527,7 +533,7 @@ makeXLFD(char *filename, FT_Face face, int isBitmap)
         t1info = &t1info_rec;
     else
         t1info = NULL;
-        
+
     if(!family)
         family = getName(face, TT_NAME_ID_FONT_FAMILY);
     if(!family)
@@ -565,7 +571,7 @@ makeXLFD(char *filename, FT_Face face, int isBitmap)
             }
         }
     }
-            
+
     if(t1info) {
         if(!family)
             family = t1info->family_name;
@@ -680,11 +686,7 @@ readFontScale(HashTablePtr entries, char *dirname)
     char *filename;
     FILE *in;
     int rc, count, i;
-    char file[MAXFONTFILENAMELEN], font[MAXFONTNAMELEN];
-    char format[100];
-
-    snprintf(format, 100, "%%%ds %%%d[^\n]\n", 
-             MAXFONTFILENAMELEN, MAXFONTNAMELEN);
+    char file[MAXFONTFILENAMELEN+1], font[MAXFONTNAMELEN+1];
 
     if(dirname[n - 1] == '/')
         filename = dsprintf("%sfonts.scale", dirname);
@@ -709,7 +711,10 @@ readFontScale(HashTablePtr entries, char *dirname)
     }
 
     for(i = 0; i < count; i++) {
-        rc = fscanf(in, format, file, font);
+        rc = fscanf(in,
+		    "%" STRINGIFY(MAXFONTFILENAMELEN) "s "
+		    "%" STRINGIFY(MAXFONTNAMELEN) "[^\n]\n",
+		    file, font);
         if(rc != 2)
             break;
         putHash(entries, font, file, 100);
@@ -752,7 +757,7 @@ filePrio(char *filename)
 }
 
 static int
-doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
+doDirectory(const char *dirname_given, int numEncodings, ListPtr encodingsToDo)
 {
     char *dirname, *fontscale_name, *filename, *encdir;
     FILE *fontscale, *encfile;
@@ -782,9 +787,9 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
         exit(1);
     }
 
-    if (onlyEncodings) 
+    if (onlyEncodings)
 	goto encodings;
-    
+
     entries = makeHashTable();
     if(doBitmaps && !doScalable) {
         readFontScale(entries, dirname);
@@ -824,6 +829,9 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
     while((entry = readdir(dirp)) != NULL) {
         int have_face = 0;
         char *xlfd_name = NULL;
+	struct stat f_stat;
+	int tprio = 1;
+
         xlfd = NULL;
 
 	if (xl) {
@@ -834,6 +842,23 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
 
         filename = dsprintf("%s%s", dirname, entry->d_name);
 
+#define PRIO(x) ((x << 1) + tprio)
+#ifdef DT_LNK
+	if (entry->d_type != DT_UNKNOWN) {
+	    if (entry->d_type == DT_LNK)
+		tprio = 0;
+	} else
+#endif
+#ifdef S_ISLNK
+	{
+	    if (lstat(filename, &f_stat))
+		goto done;
+	    if (S_ISLNK(f_stat.st_mode))
+		tprio = 0;
+	}
+#else
+	;
+#endif
         if(doBitmaps)
             rc = bitmapIdentify(filename, &xlfd_name);
         else
@@ -860,7 +885,7 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
                         isBitmap = 1;
                 }
             }
-        
+
             if(isBitmap) {
                 if(!doBitmaps)
                     goto done;
@@ -873,10 +898,9 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
                 BDF_PropertyRec prop;
                 rc = FT_Get_BDF_Property(face, "FONT", &prop);
                 if(rc == 0 && prop.type == BDF_PROPERTY_TYPE_ATOM) {
-                    xlfd_name = malloc(strlen(prop.u.atom) + 1);
+                    xlfd_name = strdup(prop.u.atom);
                     if(xlfd_name == NULL)
                         goto done;
-                    strcpy(xlfd_name, prop.u.atom);
                 }
             }
         }
@@ -894,7 +918,7 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
                 xlfd = listCons(s, xlfd);
             } else {
                 /* Not a reencodable font -- skip all the rest of the loop body */
-                putHash(entries, xlfd_name, entry->d_name, filePrio(entry->d_name));
+                putHash(entries, xlfd_name, entry->d_name, PRIO(filePrio(entry->d_name)));
                 goto done;
             }
         }
@@ -928,27 +952,26 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
                     found = 1;
                     snprintf(buf, MAXFONTNAMELEN, "%s-%s",
                             lp->value, encoding->value);
-                    putHash(entries, buf, entry->d_name, filePrio(entry->d_name));
+                    putHash(entries, buf, entry->d_name, PRIO(filePrio(entry->d_name)));
                 }
             }
-            for(encoding = extra_encodings; encoding; 
+            for(encoding = extra_encodings; encoding;
                 encoding = encoding->next) {
                 if(checkExtraEncoding(face, encoding->value, found)) {
                     /* Do not set found! */
                     snprintf(buf, MAXFONTNAMELEN, "%s-%s",
                             lp->value, encoding->value);
-                    putHash(entries, buf, entry->d_name, filePrio(entry->d_name));
+                    putHash(entries, buf, entry->d_name, PRIO(filePrio(entry->d_name)));
                 }
             }
         }
     done:
-        if(have_face) {
+        if(have_face)
             FT_Done_Face(face);
-            have_face = 0;
-        }
         deepDestroyList(xlfd);
         xlfd = NULL;
         free(filename);
+#undef PRIO
     }
 
     closedir(dirp);
@@ -980,6 +1003,7 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
 	    exit(1);
 	}
         fprintf(encfile, "%d\n", numEncodings);
+        encodingsToDo = sortList(encodingsToDo);
         for(lp = encodingsToDo; lp; lp = lp->next) {
             fprintf(encfile, "%s\n", lp->value);
         }
@@ -993,7 +1017,7 @@ doDirectory(char *dirname_given, int numEncodings, ListPtr encodingsToDo)
 #define CODE_IGNORED(c) ((c) < 0x20 || \
                          ((c) >= 0x7F && (c) <= 0xA0) || \
                          (c) == 0xAD || (c) == 0xF71B)
-   
+
 static int
 checkEncoding(FT_Face face, char *encoding_name)
 {
@@ -1009,7 +1033,7 @@ checkEncoding(FT_Face face, char *encoding_name)
     /* An encoding is ``small'' if one of the following is true:
          - it is linear and has no more than 256 codepoints; or
          - it is a matrix encoding and has no more than one column.
-       
+
        For small encodings using Unicode indices, we require perfect
        coverage except for CODE_IGNORED and KOI-8 IBM-PC compatibility.
 
@@ -1024,8 +1048,8 @@ checkEncoding(FT_Face face, char *encoding_name)
             if(mapping->type == FONT_ENCODING_POSTSCRIPT) {
                 if(encoding->row_size > 0) {
                     for(i = encoding->first; i < encoding->size; i++) {
-                        for(j = encoding->first_col; 
-                            j < encoding->row_size; 
+                        for(j = encoding->first_col;
+                            j < encoding->row_size;
                             j++) {
                             n = FontEncName((i<<8) | j, mapping);
                             if(n && FT_Get_Name_Index(face, n) == 0) {
@@ -1051,12 +1075,12 @@ checkEncoding(FT_Face face, char *encoding_name)
         if(find_cmap(mapping->type, mapping->pid, mapping->eid, face)) {
             int total = 0, failed = 0;
             if(encoding->row_size > 0) {
-                int estimate = 
+                int estimate =
                     (encoding->size - encoding->first) *
                     (encoding->row_size - encoding->first_col);
                 for(i = encoding->first; i < encoding->size; i++) {
-                    for(j = encoding->first_col; 
-                        j < encoding->row_size; 
+                    for(j = encoding->first_col;
+                        j < encoding->row_size;
                         j++) {
                         c = FontEncRecode((i<<8) | j, mapping);
                         if(CODE_IGNORED(c)) {
@@ -1111,7 +1135,7 @@ checkEncoding(FT_Face face, char *encoding_name)
     return 0;
 }
 
-static int 
+static int
 find_cmap(int type, int pid, int eid, FT_Face face)
 {
     int i, n, rc;
@@ -1134,7 +1158,7 @@ find_cmap(int type, int pid, int eid, FT_Face face)
         /* prefer Microsoft Unicode */
         for(i=0; i<n; i++) {
             cmap = face->charmaps[i];
-            if(cmap->platform_id == TT_PLATFORM_MICROSOFT && 
+            if(cmap->platform_id == TT_PLATFORM_MICROSOFT &&
                cmap->encoding_id == TT_MS_ID_UNICODE_CS) {
                 rc = FT_Set_Charmap(face, cmap);
                 if(rc == 0)
@@ -1208,8 +1232,8 @@ checkExtraEncoding(FT_Face face, char *encoding_name, int found)
     }
 }
 
-static char*
-notice_foundry(char *notice)
+static const char*
+notice_foundry(const char *notice)
 {
     int i;
     for(i = 0; i < countof(notice_foundries); i++)
@@ -1219,7 +1243,7 @@ notice_foundry(char *notice)
 }
 
 static int
-vendor_match(signed char *vendor, char *vendor_string)
+vendor_match(const signed char *vendor, const char *vendor_string)
 {
     /* vendor is not necessarily NUL-terminated. */
     int i, len;
@@ -1232,8 +1256,8 @@ vendor_match(signed char *vendor, char *vendor_string)
     return 1;
 }
 
-static char*
-vendor_foundry(signed char *vendor)
+static const char*
+vendor_foundry(const signed char *vendor)
 {
     int i;
     for(i = 0; i < countof(vendor_foundries); i++)
@@ -1266,7 +1290,7 @@ readEncodings(ListPtr encodings, char *dirname)
             closedir(dirp);
             return -1;
         }
-        
+
         names = FontEncIdentify(fullname);
         if(!names)
             continue;
@@ -1283,7 +1307,7 @@ readEncodings(ListPtr encodings, char *dirname)
                 encodingsToDo = listConsF(encodingsToDo, "%s %s", *name, n);
                 free(n);
             } else {
-                encodingsToDo = 
+                encodingsToDo =
                     listConsF(encodingsToDo, "%s %s", *name, fullname);
             }
             if(encodingsToDo == NULL) {
